// This file is part of OpenCV project.
// It is subject to the license terms in the LICENSE file found in the top-level directory
// of this distribution and at http://opencv.org/license.html

#ifndef __OPENCV_SIGNAL_PRECOMP__
#define __OPENCV_SIGNAL_PRECOMP__

#include <opencv2/core.hpp>

#endif
