#ifndef __OPENCV_PERF_PRECOMP_HPP__
#define __OPENCV_PERF_PRECOMP_HPP__

#include "opencv2/ts.hpp"
#include "opencv2/mcc.hpp"

namespace opencv_test
{
using namespace cv::mcc;
using namespace cv::ccm;
using namespace perf;
}

#endif
